/*
** 
** main.c
** (c) 2006 druidbartek
** (c) 2010 NuukiaWorld
** ====================
** Version - 1.40
**
*/

#ifndef TABLE_POS
// Display layout data for 84-pixel display
// Specifying it compile-time allows the optimization of the machine code
#define TABLE_POS(X) ("\x03\x0A\x0B_\x04\x04\x0E\x27\x0E\x07\x04\x07\x2A\x09\x01"[X])
#endif

#include "main.h"

/**************************************************************************************************/

static const unsigned char elem_list[9][18] = {
   { 1, 0, 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  2},
   { 3, 4, 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  5,  6,  7,  8,  9, 10},
   {11,12, 0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 13, 14, 15, 16, 17, 18},
   {19,20,21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36},
   {37,38,39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54},
   {55,56, 0, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86},
   {87,88, 0,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118},
   { 0, 0, 0, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71},
   { 0, 0, 0, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99,100,101,102,103}
};

static const nokstr_t *const str_tips[9] = {
   &str_mass, &str_melt_p, &str_boil_p,
   &str_density, &str_discovery, &str_configuration,
   &str_ionization, &str_states, &str_elektroneg
};

/**************************************************************************************************/

/* Find specified record in specified data block. Record end markers differ:
   First byte of record:   Record type:   Record end marker:
   '\x00'                  8-bit          Next '\x00'
   '\x04'                  UCS-2          '\x0000'L
   '\x20'...'\xFF'         ASCII          Last character has bit 7 set
   anything else           backreference  Record has only this one byte
*/
static const char *find_data( const char *data, unsigned int index ) {
   while( index != 0 ) {
      if( *data >= '\x20' ) {          // ASCII string
         while( ( *data++ & 0x80 ) == 0 );
      } else if( *data == '\x04' ) {   // UCS-2 string
         while( ( *data++ | *data++ ) != '\0' );
      } else if( *data++ == '\x00' ) { // 8-bit string
         while( *data++ != '\0' );
      }                                // Backreference
      index--;
   }
   return data;
}

/* Return specified record from current block pointed by app.data.
   The returned string has NUL at the end. Only ASCII records use the specified destination buffer.
*/
static const char *get_data( char *dest, unsigned int index ) {
   const char *data;
   char *s;
   
   do {
      data = find_data( app.data, index );
      if( *data >= '\x20' ) {          // ASCII string
         s = dest;
         do {
            *s++ = *data & 0x7F;
         } while( ( *data++ & 0x80 ) == 0 );
         *s = '\0';
         return dest;
      }
      index = (unsigned int)*data;     // Resolve backreference
   } while( *data != '\x04' && *data++ != '\x00' );
   return (const char*)data;           // UCS-2 or 8-bit string
}

/* Formats specified info string for current element.
*/
static const nokstr_t *get_info( unsigned int index ) {
   char *s;

   if( app.state == STATE_TIP ) return str_tips[index - 1];

   s = app.buf;
   if( index == 6 && app.y != 0 ) { // el. config.
      strncpy( s, config_prefix[ app.y - 1 ], 4 );
      s += 4;
   }
   
   get_data( s, index );  // get data record 1...9

   if( app.buf[0] != '-' ) {
      switch(index) {
         case 1:     // mass
            strcat( s, "u" );
            break;
         case 2:     // melting point
         case 3:     // boiling point
            strcat( s, "K" );
            break;
         case 4:     // density
            strcat( s, "g/cmE" );   // E==''
            break;
         case 7:     // ioniz. en.
            strcat( s, "eV" );
            break;
      }
   }
   return (const nokstr_t*)app.buf;
}

/* Prints string centered at x
*/
static void print_center( const nokstr_t *text, unsigned int x, unsigned int y ) {
   x -= own_get_string_width( text, app.font ) / 2;
   own_print_string( text, x, y, app.font, DM_SET );
}

static void draw_panel(void) {

   unsigned int px, py, i;
   char *s;

   freegfx_begin(true);
   switch(app.state) {
      case STATE_MAP:
         ntoa( app.number, app.buf );
         s = strchr( app.buf, '\0' );
         *s++ = '-';
         get_data( s, 0 ); // Get symbol of current element (data record 0)

	      own_draw_bitmap( &bmap_table, TABLE_POS(4), TABLE_POS(5), DM_SET ); // 0, 6
         print_center( (const nokstr_t*)app.buf, TABLE_POS(6) + TABLE_POS(7) / 2, TABLE_POS(8) );

         px = TABLE_POS(9) + (TABLE_POS(10)*app.x); // 4 , 5
         py = TABLE_POS(11) + (TABLE_POS(10)*app.y); // 10, 5
         if( app.y > 6 ) py++;
	      own_draw_box( px, py, TABLE_POS(14), TABLE_POS(14), DM_SET );
         break;

      case STATE_INF:
      case STATE_TIP:
	      own_draw_bitmap( &bmap_info, 0, 0, DM_SET );
         print_center( (const nokstr_t*)get_data( app.buf, app.lng ), TABLE_POS(12), TABLE_POS(13) );

         for( i = 1; i < 4; i++ ) {
            own_print_string( get_info( app.pos + i ), TABLE_POS(0), TABLE_POS(1) + TABLE_POS(2) * i,
               ( app.state == STATE_TIP ? app.font : &own_font ), DM_SET );
         }
         break;
   }
	freegfx_end(false);
}

/**************************************************************************************************/

static void change_element( unsigned int key ) {
   unsigned int number;

   if( key == KEY_5 || key == KEY_NAVI ) {
      switch(app.state) {
         case STATE_MAP:
            app.state++; // = STATE_INF;
            break;
         case STATE_INF:
            if( ++app.lng >= records_per_element ) app.lng = 10;
            break;
      }

   } else if( key >= KEY_1 && key <= KEY_9 ) {
      do {
         if( key >= KEY_1 && key <= KEY_3 ) {
            if( app.y-- == 0 ) app.y = 8;
         }
         if( key >= KEY_7 && key <= KEY_9 ) {
            if( app.y++ >= 8 ) app.y = 0;
         }
         if( key == KEY_1 || key == KEY_4 || key == KEY_7 ) {
            if( app.x-- == 0 ) app.x = 17;
         }
         if( key == KEY_3 || key == KEY_6 || key == KEY_9 ) {
            if( app.x++ >= 17 ) app.x = 0;
         }
         number = elem_list[app.y][app.x];
      } while( number == 0 );
      app.number = number;
      app.data = find_data( element_data, ( number - 1 ) * records_per_element );
   }
   
   if( key == KEY_HASH ) {
      switch(app.state) {
         case STATE_MAP:
            send_message( MSG_START_HANDLER | MSG_2P, id_game_instructions, str_instr_text );
//            app.state = STATE_HLP;
            break;
         case STATE_INF:
            app.state++; // = STATE_TIP;
            break;
      }
   }
}

/**************************************************************************************************/

extern rexx_uint main_dispatcher_id; // dispatcher's call id
rexx("main_dispatcher_id=add_5e0_id(main_dispatcher,locate('action_select_back'))"); // auto creation of the id

unsigned int main_dispatcher( unsigned int msg ) {

   switch(msg) {
      case MSG_D_INIT:
         app.state = STATE_MAP;
         app.x = app.y = app.pos = 0;
         app.lng = 10;              // Default to first language
         app.number = 1;            // Start from Hydrogen
         app.data = element_data;   // Hydrogen data is here
         app.font = own_get_font("small/plain");
         
   	   freegfx_open();
   		draw_panel();
   		// FALL-THROUGH
      case MSG_D_QUIT_NOW: // called when we get back from help dialog
   		msg = MSG_R_STAY;
   		break;

      case MSG_D_QUIT:
         freegfx_close();
         msg = MSG_R_QUIT;
         break;

      case MSG_BACK:
         if( app.state != STATE_MAP ) {
            msg = MSG_NOP;
            app.state = STATE_MAP;
            draw_panel();
         }
         break;

      case MSG_KEYPRESS:
         change_element(own_get_unify_keycode(UKM_NONE));
         draw_panel();
   		break;

   	case MSG_KEYRELEASE:
         if( app.state == STATE_TIP ) {
            app.state--;
            draw_panel();
         }
         break;

      case MSG_KEY_DOWN:
         if( app.pos >= 6 ) {
            app.pos = 0;
         } else {
            app.pos += 3;
         }
         draw_panel();
         break;
      
      case MSG_KEY_UP:
         if( app.pos == 0 ) {
            app.pos = 6;
         } else {
            app.pos -= 3;
         }
         draw_panel();
         break;

      case MSG_KEYPRESS_CHAR:
         msg = MSG_NOP;
         break;
   }
   return msg;
}
